﻿/*
 * 24 Puzzle
 * Copyright (C) 2008 Michael Birken
 * 
 * This file is part of 24 Puzzle.
 *
 * 24 Puzzle is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published 
 * by the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * 24 Puzzle is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace TwentyFourPuzzle {

  public interface IElement {
    void Evaluate(Stack<double> stack);
    void Print(Stack<string> stack);
  }

  public class Number : IElement {

    public double value;

    public Number(double value) {
      this.value = value;
    }
    
    public void Evaluate(Stack<double> stack) {
      stack.Push(value);
    }

    public void Print(Stack<string> stack) {
      stack.Push(value.ToString());
    }
  }

  public delegate double UnaryFunction(double a);
  public delegate double BinaryFunction(double a, double b);

  public class UnaryOperator : IElement {

    private UnaryFunction unaryFunction;
    private string name;
    private bool before;

    public UnaryOperator(
        UnaryFunction unaryFunction, string name, bool before) {
      this.unaryFunction = unaryFunction;
      this.name = name;
      this.before = before;
    }
    
    public void Evaluate(Stack<double> stack) {
      stack.Push(unaryFunction(stack.Pop()));
    }
    
    public void Print(Stack<string> stack) {
      stack.Push(string.Format(
          before ? "{0}({1})" : "({1}){0}", name, stack.Pop()));
    }
  }

  public class BinaryOperator : IElement {

    private BinaryFunction binaryFunction;
    private string name;

    public BinaryOperator(BinaryFunction binaryFunction, string name) {
      this.binaryFunction = binaryFunction;
      this.name = name;
    }
    
    public void Evaluate(Stack<double> stack) {
      double b = stack.Pop();
      double a = stack.Pop();
      stack.Push(binaryFunction(a, b));
    }
    
    public void Print(Stack<string> stack) {
      string b = stack.Pop();
      string a = stack.Pop();
      stack.Push(string.Format("({0} {1} {2})", a, name, b));
    }
  }
  
  public class Example4 {

    private void Solve(
        Number[] numbers, 
        UnaryOperator[] unaryOperators, 
        BinaryOperator[] binaryOperators, 
        double target,
        List<IElement> elements,
        bool[] numbersAvailable,
        bool[] unaryOperatorsAvailable,
        int stackSize,
        Stack<double> numberStack,
        Stack<string> stringStack) {

      bool remainingNumbers = false;
      double lastNumber = Double.NaN;
      for (int i = 0; i < numbersAvailable.Length; i++) {
        if (numbersAvailable[i] && numbers[i].value != lastNumber) {
          remainingNumbers = true;
          elements.Add(numbers[i]);
          numbersAvailable[i] = false;
          Solve(numbers, unaryOperators, binaryOperators, target, elements,
              numbersAvailable, unaryOperatorsAvailable, stackSize + 1,
              numberStack, stringStack);
          numbersAvailable[i] = true;
          elements.RemoveAt(elements.Count - 1);
        }
      }

      if (stackSize >= 1) {
        for (int i = 0; i < unaryOperatorsAvailable.Length; i++) {
          if (unaryOperatorsAvailable[i]) {
            elements.Add(unaryOperators[i]);
            unaryOperatorsAvailable[i] = false;
            Solve(numbers, unaryOperators, binaryOperators, target, elements,
                numbersAvailable, unaryOperatorsAvailable, stackSize,
                numberStack, stringStack);
            unaryOperatorsAvailable[i] = true;
            elements.RemoveAt(elements.Count - 1);
          }
        }
      }

      if (stackSize >= 2) {
        for (int i = 0; i < binaryOperators.Length; i++) {
          elements.Add(binaryOperators[i]);
          Solve(numbers, unaryOperators, binaryOperators, target, elements,
              numbersAvailable, unaryOperatorsAvailable, stackSize - 1,
              numberStack, stringStack);
          elements.RemoveAt(elements.Count - 1);
        }
      }

      if (stackSize == 1 && !remainingNumbers) {
        if (target == EvalutateExpression(elements, numberStack)) {
          PrintExpression(elements, target, stringStack);
        }
      }
    }

    private double EvalutateExpression(List<IElement> elements, 
        Stack<double> stack) {
      foreach (IElement element in elements) {
        element.Evaluate(stack);
      }
      return stack.Pop();
    }

    private void PrintExpression(List<IElement> elements, double target,
        Stack<string> stack) {
      foreach (IElement element in elements) {
        element.Print(stack);
      }
      string expression = stack.Pop();
      Console.WriteLine("{0} = {1}", expression, target);
    }

    public void Solve(double[] numbers, UnaryOperator[] unaryOperators, 
        BinaryOperator[] binaryOperators, double target) {
      Array.Sort(numbers);
      Number[] nums = new Number[numbers.Length];
      for (int i = 0; i < numbers.Length; i++) {
        nums[i] = new Number(numbers[i]);
      }
      List<IElement> elements = new List<IElement>();
      bool[] numbersAvailable = new bool[numbers.Length];
      for (int i = 0; i < numbersAvailable.Length; i++) {
        numbersAvailable[i] = true;
      }
      bool[] unaryOperatorsAvailable = new bool[unaryOperators.Length];
      for (int i = 0; i < unaryOperatorsAvailable.Length; i++) {
        unaryOperatorsAvailable[i] = true;
      }
      Stack<double> numberStack = new Stack<double>();
      Stack<string> stringStack = new Stack<string>();
      Solve(nums, unaryOperators, binaryOperators, target, elements, 
          numbersAvailable, unaryOperatorsAvailable, 0, numberStack, 
          stringStack);
    }
  }

  
}
